#!/usr/local/bin/perl
#
# $Source: /tmp_mnt/net/dworshak2/users/student/evans911/prog/cs481/src/RCS/assign_position.pl,v $
# $Author: evans911 $
# Current revision: $Revision: 1.20 $
# Last modified: $Date: 1996/11/13 02:00:36 $
# State of current revision: $State: Exp $
# Currently locked by: $Locker:  $
#
# Description: These routines calculate the locations of nodes and lines
#              on the graph.  This information is stored in the last four
#              fields of the stack.
#              
# $Log: assign_position.pl,v $
# Revision 1.20  1996/11/13 02:00:36  evans911
# Changed all element offsets that went from the end of an item to
# calculate offsets from the beginning of the item.
#
# Revision 1.19  1996/11/12 23:46:31  evans911
# Cleaned up and updated comments.
#
# Revision 1.18  1996/11/08 00:34:27  evans911
# Seriously hacked to pass the stack by reference.
#
# Revision 1.17  1996/10/30 21:56:37  suthe922
# fixing comments
#
# Revision 1.16  1996/10/25 06:49:14  suthe922
# *** empty log message ***
#
# Revision 1.15  1996/10/23 21:11:17  evans911
# Modified assign_position() to check for an empty stack before
# calling y_position, so that a file with V(g) == 1 graphs properly.
#
# Revision 1.14  1996/10/16 23:51:39  suthe922
# Just a small amount of cleaning
#
# Revision 1.13  1996/10/10 15:27:43  suthe922
# Added a tenth field to the stack elements (blk_x or block x) that
# designates the position to draw a horizontal line to from each
# branching node. Also added statement in physical_x_position() to
# insert value into that field.
#
# Revision 1.12  1996/10/09 22:14:00  evans911
# Added author line to sub headers and deleted some commented out
# debugging code.
#
# Revision 1.11  1996/10/03 07:08:23  suthe922
# Took out debug spew and added real comments
#
# Revision 1.10  1996/10/02 05:53:19  suthe922
# fixed max_x counter in physical_x_position
#
# Revision 1.9  1996/10/02 04:15:19  suthe922
# physical_x_position appears to work
#
# Revision 1.8  1996/10/02 03:51:09  suthe922
# just got rid of some debug spew
#
# Revision 1.7  1996/10/02 03:45:57  suthe922
# logical_x_position and y_position now both work
# going on to physical_x_position
#
# Revision 1.6  1996/10/02 00:26:15  suthe922
# more progress on y_position
# currently working on function alt_y_position so
# I don't erase the original until I have this one
# working
#
# Revision 1.5  1996/10/01 23:32:45  suthe922
# logical_x_position is working
# starting on y_position
#
# Revision 1.4  1996/09/30 06:53:59  suthe922
# working on fixing problems in y_position -
# starting on different idea, finding logical x first,
# then use that to find the y
#
# Revision 1.3  1996/09/30 05:48:44  suthe922
# took out garbage from y_position
#
# Revision 1.2  1996/09/30 05:17:29  suthe922
# y_position works - don't touch it!!!!!!!!
#
# Revision 1.1  1996/09/30 03:35:44  suthe922
# Initial revision
#

#
# Subroutine: assign_position()
# Arguments:     $data_size : number of items in a stack element
#                $stack = reference to the stack:
#                      #ppd : preprocessor directive
#                      expr : expression for #ppd
#                      lin# : line number
#                      blk_x: distance over from p_x to draw line from #ppd
#                      p_y    : physical y position for #ppd
#                      l_x   : logical x for #ppd
#                      p_x   : physical x for #ppd
#            
# Return Values: $max_x     : the maximum x for any part of the graph
#                $max_y     : the maximum y for the graph (location of the 
#                             last #endif
#
# Globals Used: none
#               
# Required Files: none
#                 
# Author: Noah Sutherland
#
# Description: calls functions to determine the x and y positions on the graph
#              for each #directive and the maximum x and y for the graph
#         
sub assign_position{
    my ($data_size, $stack) = @_;
    my (@sublist) = ();
    my ($i, $j,
	$stack_size,
	$max_lx, $max_x, $max_y,
	$next_element);

    # find the logical x for each element by calling logical_x_position()
    ($max_lx) = &logical_x_position(-1,
				    $data_size,
				    $stack);

    if ($#$stack == -1)
    {
	# stack is empty
	$max_y = 0;
    }
    else
    {
	# find the y for each element by calling y_position()
	($max_y, $next_element) = &y_position(1,
					      0,
					      $data_size,
					      $stack);
    }

    # find the physical x for each element by calling physical_x_position()
    ($max_x) = &physical_x_position(0, 
				    0, 
				    $data_size,
				    $stack);

    #return the max values
    return($max_x, $max_y);
}

#
# Subroutine: y_position()
# Arguments:     $current_y : the y value for the next directive
#                $current_lx: the logical x value for this recursion step
#                $data_size : the number of items in a stack element
#                $stack     : reference to the stack starting with the next 
#                             element to be looked at 
#
# Return Values: $max_y     : the highest y value seen on this recursion step
#                $i         : index into @stack of the next element to look at
#
# Globals Used: none
#               
# Required Files: none
#                 
# Author: Noah Sutherland
#
# Description: Look at each element in the stack at the current logical level.
#              If it is not an endif, recurse on elements of stack within
#              branch. Set the y value for each element and return if you
#              reach the end of the stack or an element at a lower logical
#              level.
#         
sub y_position{
    my ($current_y, $current_lx, $data_size, $stack) = @_;
    my ($max_y, $block_max_y) = 1;
    my ($i);
    my (@new_stack) = ();
    
    # loop for each element in the stack
    for($i = 0; $i < $#$stack + 1; $i = $i + $data_size){

	# increment max_y if needed
	if($current_y > $max_y){
	    $max_y = $current_y;
	}

	# check if the current element is on the current logical level
	if($$stack[$i + 5] == $current_lx){

	    # if it is not an #endif
	    if($$stack[$i] !~ m/\#endif/){
		# set the y value
		$$stack[$i + 4] = $current_y;
		# if the next element is in a deeper logical level, recurse
		# with stack starting at the current element
		if($$stack[$i + $data_size + 5] == ($current_lx + 1)){
		    @new_stack = splice(@$stack, 0, $i + $data_size);
		    ($block_max_y, 
		     $i) = y_position($current_y + 1, 
					  $current_lx + 1, 
					  $data_size, 
					  $stack);
		    # update max_y if needed
		    if($block_max_y > $max_y){
			$max_y = $block_max_y;
		    }
		    # put the branch evaluated back into the whole stack
		    $i = @new_stack + $i - $data_size;
		    @$stack = (@new_stack, @$stack);
		}
		$current_y++;
	    }
	    # if the current element is an #endif, set its y value
	    else{
		if($max_y > $current_y){
		    $current_y = $max_y;
		}
		$$stack[$i + 4] = $current_y;
		$current_y++;
	    }
	}
	# if the current element is at a lower logical level, return
	elsif($$stack[$i + 5] < $current_lx){
	    return($max_y, $i);
	}
    }
    # if we reach the end of the stack, return
    return($max_y, $i);
}

#
# Subroutine: logical_x_position()
# Arguments:     $current_x : the current logical x
#                $data_size : # of items in a stack element
#                $stack     : reference to the stack of directives
#
# Return Values: $max_x     : the maximum logical x value found
#
# Globals Used: none
#               
# Required Files: none
#                 
# Author: Noah Sutherland
#
# Description: Look at each element in the stack. Logical x becomes greater 
#              each time an if is encountered and smaller at each endif.
#              Each element is assigned a logical x value and the stack is
#              returned.
#
sub logical_x_position{
    my ($current_x, $data_size, $stack) = @_;
    my ($i, $max_x) = (0, 0, 0, 0);

    # loop for each element in the stack
    for($i = 0; $i < @$stack; $i += $data_size){

	# if the current element is an endif, assign x then decrement x
	if($$stack[$i] =~ m/\#endif/){
	    $$stack[$i + 5] = $current_x;
	    $current_x--;
	}
	# if the current element is an elif or an else, assign x
	elsif($$stack[$i] =~ m/\#elif|\#else/){
	    $$stack[$i + 5] = $current_x;
	}
	# if the current element is any of the ifs, increment x and assign it
	else{
	    $current_x++;
	    $$stack[$i + 5] = $current_x;
	}

	# increment max_x if needed
	if($current_x > $max_x){
	    $max_x = $current_x;
	}
    }
    # when we have gone through the whole stack, return
    return($max_x);
}

#
# Subroutine: physical_x_position()
# Arguments:   $current_p_x : the current physical x position
#              $current_l_x : the current logical x position
#              $data_size   : # of items in a stack element
#              $stack       : reference to the stack of directives containing 
#                             logical x values
#
# Return Values: $max_x     : the maximum physical x value seen at the current
#                             logical level
#
# Globals Used: none
#               
# Required Files: none
#                 
# Author: Noah Sutherland
#
# Description: Look at each element in the stack. Find a logical block.
#              Recurse on each branch in that block. Assign x values to
#              each element in the stack and return.
#         
sub physical_x_position{
    my ($current_p_x, $current_l_x, $data_size, $stack) = @_;
    my ($i, $j, $k, $max_x, $block_x);
    my (@block, @block_nodes, @rest) = ();

    $max_x = $block_x = 0;

    # loop for each element in the stack
    for($i = 0; $i < $#$stack + 1; $i += $data_size){

	# make a list, @block_nodes, of the elements in the stack at the 
	# current logical level that make up the next block (if - endif)
	$block_x = $current_p_x;
	@block_nodes = ();
	until(($$stack[$i] =~ m/\#endif/) 
	      && ($$stack[$i + 5] == $current_l_x)){
	    if($$stack[$i + 5] == $current_l_x){
		@block_nodes = (@block_nodes, $i);
	    }
	    $i += $data_size;
	}
	
	$$stack[$i + 6] = $current_p_x;
	$k = $i;

	# count down to find each element listed in block_nodes
	for($j = $k - $data_size; 
	    $j >= $block_nodes[0]; 
	    $j -= $data_size){

	    # assign the position to draw the horizontal line to
	    if($$stack[$j] !~ m/\#endif/){
		$$stack[$j + 3] = $block_x + 1;
	    }
	    # for each branch in the block
	    if($$stack[$j + 5] == $current_l_x){
		$$stack[$j + 6] = $current_p_x;
		@rest = splice(@$stack, $k);
		@block = splice(@$stack, $j + $data_size);
		# if the branch is not empty, recurse on the elements in 
		# the branch
		if($k > $j + $data_size){
		    ($block_x) = &physical_x_position($block_x + 1, 
							      $current_l_x + 1,
							      $data_size, 
							      \@block);
		}
		# if the block is empty, the x is incremented by 1
		elsif($$stack[$j] !~ m/\#endif/){
		    $block_x++;
		}
		@$stack = (@$stack, @block, @rest);
		$k = $j;
	    }
	}
	# increment max_x if needed
	if($block_x > $max_x){
	    $max_x = $block_x;
	}
    }
    # return after looking at the whole stack
    return($max_x);
}

1;
