#!/usr/local/bin/perl -w
#
# $Source: /tmp_mnt/net/dworshak2/users/student/evans911/prog/cs481/src/RCS/gen_image.pl,v $
# $Author: evans911 $
# Current revision: $Revision: 0.17 $
# Last modified: $Date: 1996/11/13 01:52:15 $
# State of current revision: $State: Exp $
# Currently locked by: $Locker:  $
#
# Description: The functions in this file use the PS api to generate the 
#              graph.
#              
# $Log: gen_image.pl,v $
# Revision 0.17  1996/11/13 01:52:15  evans911
# Changed all element offsets that went from the end of an item to
# calculate offsets from the beggining of the item.
#
# Revision 0.16  1996/11/13 00:19:59  evans911
# Cleaned up comments.
#
# Revision 0.15  1996/11/11 20:45:43  evans911
# Deleted a line of code that was multiplying the scale factor for
# one-page scaling by 4.
#
# Revision 0.14  1996/11/11 19:14:08  evans911
# Made modifications to add two scale factors.
#
# Revision 0.13  1996/11/08 00:31:51  evans911
# Modified to pass the stack by reference.
#
# Revision 0.12  1996/11/07 22:20:37  evans911
# Fixed file name on printed output.
# Fixed boundary case where an extra row sometimes get printed.
#
# Revision 0.11  1996/11/07 20:04:26  evans911
# Modified calls to PS*() functions to handle the new way of doing
# things (associative array in memory, pass by reference).
#
# Revision 0.10  1996/10/25 17:53:39  evans911
# Now prints LOC in all the appropriate places.
# Also fixed scaling for graphs with no #CCD logic.
#
# Revision 0.9  1996/10/24 17:02:46  evans911
# Added comments.
#
# Revision 0.8  1996/10/23 21:15:16  evans911
# Made modifications to y coordinate that reflect the changes I made
# in assign_position.pl.  Now that the y position is larger by 1,
# there is no need to add 1 every time an object is drawn.
#
# Revision 0.7  1996/10/17 20:46:27  evans911
# This one works, and is not tail recursive.
#
# Revision 0.6  1996/10/17 15:12:11  evans911
# Fixed the LOC problems.  The first BLOC is still incorrect, since there
# is no place to get this number, except by calculating it.  Also,
# the tail recursion still remains.
#
# Revision 0.5  1996/10/16 22:43:50  evans911
# It actually generates a correct graph now!  There are still some
# problems with LOC though.  In addition, draw_block() is tail
# recursive, which causes deep recursion warnings for large files.
#
# Revision 0.4  1996/10/16 01:50:33  evans911
# Changed $scale_factor check from 0 for fit to page to 'p'.
#
# Revision 0.3  1996/10/16 01:23:46  evans911
# Well, this version works for the mockup sample, but has a number of
# problems still, including:
# * Multiple blocks at the same depth within another block causes
#   some weird problems with the lines.  It looks like negative parms
#   are being sent to the PS line drawing functions.
# * LOC is incorrect in at least some places, if not all.
#
# Revision 0.2  1996/10/09 23:09:58  evans911
# Added author line to comment block.
#
# Revision 0.1  1996/09/18 02:08:06  evans911
# Initial revision.  Just an empty shell.
#
#

#
# Subroutine: gen_image(@stack)
# Arguments: $graphfile    : name of graph output file
#            $infile       : name of code input file
#            $item_size    : number of elements in one stack item
#            $max_x        : maximum x position of any graph element
#            $max_y        : maximum y position of any graph element
#            $file_loc     : total LOC for code input file
#            $scale_factor : scale factor for graph
#            $CCD_list     : reference to the stack of #CCDs
#            
# Return Values: 
#                
# Globals Used: 
#               
# Required Files: 
#                 
# Author: Jason Evans
#
# Description: Generates a PostScript file by using the PS api and 
#              draw_block().
#              
sub gen_image
{
    my ($graphfile, $infile, 
	$item_size, $max_x, $max_y, $file_loc, $scale_factor,
	$CCD_list) = @_;
    my ($curr_item) = 0;
    my ($xpages, $ypages);
    my ($xscale, $yscale);
    my (%page_block);
    my ($top_loc);
    my ($junk1, $junk2, $junk3);

    # calculate the scale factor and number of pages
    if ($scale_factor eq 'p')
    {
	# scale to fit on one page
	$xpages = 1;
	$ypages = 1;
        $xscale = 64 / ($max_x + 3);
        $yscale = 80 / ($max_y + 2);
        $scale_factor = $xscale;
        if ($scale_factor > $yscale)
        {
            $scale_factor = $yscale;
        }
    }
    else
    {
	$xpages = int (($max_x + 2) / (64 / $scale_factor)) + 1;
	$ypages = int (($max_y + 1) / (80 / $scale_factor)) + 1;
    }

    # initialize page block
    %page_block = &PS_create_page_block($infile, $xpages, $ypages);
    
    # draw initial elbow.
    if ($#$CCD_list == -1)
    {
	$top_loc = $file_loc . " LOC";
    }
    else
    {
	$top_loc = "";
    }
	&PS_draw_top_line(\%page_block,
			  $scale_factor,
			  $top_loc,
			  1,
			  0, 
			  0
			  );

    # recursive part
    &draw_block(\%page_block,
		$scale_factor,
		$xpages,
		0,
		$item_size,
		$curr_item,
		$max_y,
		$file_loc,
		0,
		0,
		$CCD_list
		);

    # draw last elbow
    &PS_draw_bottom_line(\%page_block,
			 $scale_factor,
			 1, 
			 0, 
			 $max_y + 1
			 );

    # finish up
    &PS_create_file($graphfile,
		    $scale_factor / 4,
		    $xpages,
		    $ypages,
		    \%page_block
		    );
}

#
# Subroutine: ($curr_item, $curr_x, $curr_y) = draw_block(*PSOUT,
#                                                         $depth,
#                                                         $item_size,
#                                                         $curr_item,
#                                                         $max_y,
#                                                         $file_loc,
#                                                         $curr_x,
#                                                         $curr_y,
#                                                         @CCD_list)
#                           
# Arguments: *PSOUT     : global open file descriptor used for PS output
#            $depth     : current logical depth
#            $item_size : # of elements in one list item
#            $curr_item : index of current item
#            $max_y     : max y coord. for entire PS graph
#            $file_loc  : LOC for entire input file
#            $curr_x    : current x position (used for matching up lines)
#            $curr_y    : current y position (used for matching up lines)
#            $CCD_list  : reference to the stack containing all needed #CCD
#                         info
#            
# Return Values: $curr_item : current position in list
#                $curr_x    : current x position
#                $curr_y    : current y position
#                
# Globals Used: *PSOUT
#               
# Required Files: 
#                 
# Author: Jason Evans
#
# Description: This recursive function makes calls to the PS api in ps.pl,
#              thereby inserting all the nodes and lines besides the first and
#              last elbows (done by gen_image()).  It is recursive because
#              there is no easy way to know where to start drawing vertical
#              lines to connect to the bottom lines.  This function solves
#              this problem by keeping track of the (x, y) position of what
#              was most recently drawn.  When the function returns, it returns
#              this value so that the next logical level up knows where to
#              start drawing the vertical line.
#              
sub draw_block
{
    my ($page_block, $scale_factor, $xpages, $depth, $item_size, $curr_item, 
	$max_y, $file_loc, $curr_x, $curr_y, $CCD_list) = @_;
    my($loc, $i, $j);
    my($has_no_else);
    my($temp_x, $temp_y);
    my($top_line_item, $line_loc);

    #
    # if (we're at the end of the input) 
    #    or (the current item isn't a #if, #ifdef, or #ifndef)
    # short circuit (return)
    #
    unless (($curr_item * $item_size < $#$CCD_list)
	    &&
	    (($$CCD_list[$curr_item * $item_size]) =~ /\#if/))
    {
	return ($curr_item, -1, -1);
    }

    #
    # Outer loop loops once for each block at the current logical depth.
    #
    while ((($curr_item + 1) * $item_size < $#$CCD_list)
	   && ($$CCD_list[$curr_item * $item_size + 5]
	       <= $$CCD_list[($curr_item + 1) * $item_size + 5])
	   && ($$CCD_list[$curr_item * $item_size + 5] == $depth))
    {
	$has_no_else = 1;

	#
	# Inner loop loops once for each #if..., #elif, and #else for the
	# current block.
	#
	while ((($curr_item * $item_size) < $#$CCD_list)
	       && ((($$CCD_list[$curr_item * $item_size]) !~ /\#endif/)
		   || ($$CCD_list[$curr_item * $item_size + 5] 
		       != $depth)))
	{
	    #
	    # Save item offset for top line, which can't be drawn until we know
	    # whether LOC should be printed.
	    #
	    $top_line_item = $curr_item;

	    #
	    # Increment current position to reflect the fact that we're
	    # about to recurse into a branch.
	    #
	    $curr_x = $$CCD_list[$curr_item * $item_size + 3] + 1;
	    $curr_y = $$CCD_list[$curr_item * $item_size + 4] + 2;

	    #
	    # draw current node
	    #

	    # if a (#if, #ifdef, or #ifndef)
	    if ($$CCD_list[$curr_item * $item_size] =~ /\#if/)
	    {
		if ($curr_item == 0)
		{
		    $loc = ($$CCD_list[$curr_item * $item_size + 2] - 1) 
			. " LOC";
		}
		elsif ($$CCD_list[$curr_item * $item_size + 5]
		       > $$CCD_list[($curr_item - 1) * $item_size + 5])
		{
		    $loc = ($$CCD_list[$curr_item * $item_size + 2]
			    - $$CCD_list[($curr_item - 1) * $item_size + 2] - 1)
			. " LOC";
		}
		else
		{
		    $loc = "";
		}

		&PS_insert_if($page_block,
			      $scale_factor,
			      $xpages,
			      $$CCD_list[$curr_item * $item_size],
			      $$CCD_list[$curr_item * $item_size + 1],
			      $$CCD_list[$curr_item * $item_size + 2],
			      $loc,
			      0,
			      $$CCD_list[$curr_item * $item_size 
					+ 6],
			      $$CCD_list[$curr_item * $item_size 
					+ 4]
			      );
	    }

	    # if a (#elif)
	    elsif ($$CCD_list[$curr_item * $item_size] =~ /\#elif/)
	    {
		&PS_insert_if($page_block,
			      $scale_factor,
			      $xpages,
			      $$CCD_list[$curr_item * $item_size],
			      $$CCD_list[$curr_item * $item_size + 1],
			      $$CCD_list[$curr_item * $item_size + 2],
			      "",
			      1,
			      $$CCD_list[$curr_item * $item_size 
					 + 6],
			      $$CCD_list[$curr_item * $item_size 
					 + 4]
			      );
	    }

	    # if a (#else)
	    elsif ($$CCD_list[$curr_item * $item_size] =~ /\#else/)
	    {
		&PS_insert_else($page_block,
				      $scale_factor,
				      $$CCD_list[$curr_item * $item_size + 2],
				      $$CCD_list[$curr_item * $item_size 
						+ 6],
				      $$CCD_list[$curr_item * $item_size
						+ 4]
				      );
		$has_no_else = 0;
	    }

	    #
	    # if (not at the end of input)
	    # recurse
	    #
	    if ($curr_item * $item_size < $#$CCD_list)
	    {
		# recurse on true path
		($curr_item, 
		 $temp_x, $temp_y) = &draw_block($page_block,
						 $scale_factor,
						 $xpages,
						 $depth + 1,
						 $item_size,
						 ++$curr_item,
						 $max_y,
						 $file_loc,
						 $curr_x,
						 $curr_y,
						 $CCD_list
						 );

		if ($temp_x != -1)
		{
		    $curr_x = $temp_x;
		}
		if ($temp_y != -1)
		{
		    $curr_y = $temp_y;
		}
	    }

	    #
	    # Draw top line, but first decide whether to print LOC, depending
	    # on whether the block had any nested #CCD logic.
	    #
	    if (($temp_x == -1) && ($temp_y == -1))
	    {
		$line_loc = $$CCD_list[($top_line_item + 1) * $item_size + 2]
		    - $$CCD_list[$top_line_item * $item_size + 2] - 1
			. " LOC";
			
	    }
	    else
	    {
		$line_loc = "";
	    }
	    &PS_draw_top_line($page_block,
				    $scale_factor,
				    $line_loc,
				    $$CCD_list[$top_line_item * $item_size 
					      + 3]
				    -
				    $$CCD_list[$top_line_item * $item_size
					      + 6],
				    $$CCD_list[$top_line_item * $item_size 
					      + 6]
				    + 1,
				    $$CCD_list[$top_line_item * $item_size 
					      + 4]
				    );

	    #
	    # connect line to #endif
	    #

	    # first, find the appropriate #endif
	    for ($i = $curr_item;
		 ($i * $item_size < $#$CCD_list)
		 && (($$CCD_list[$i * $item_size] !~ /\#endif/)
		     || ($$CCD_list[$i * $item_size + 5] 
			 != $depth));
		 $i++)
	    {
	    }

	    # draw the vertical line, if necessary
	    if (($i * $item_size < $#$CCD_list)
		&&
		(($$CCD_list[$i * $item_size + 4] - $curr_y + 1)
		 > 0))
	    {
		&PS_draw_vert_line($page_block,
					 $scale_factor,
					 $$CCD_list[$i * $item_size + 4]
					 - $curr_y + 1,
					 0,
					 $curr_x - 1,
					 $curr_y - 1
					 );
	    }

	    # draw the bottom line
	    if ($i * $item_size < $#$CCD_list)
	    {
		&PS_draw_bottom_line($page_block,
					   $scale_factor,
					   $curr_x - $$CCD_list[$i * $item_size
							       + 6] - 1,
					   $curr_x - 1,
					   $$CCD_list[$i * $item_size 
						     + 4]
					   );
	    }
	} # matches inner while loop

	# draw #endif
	if ($i * $item_size < $#$CCD_list)
	{
	    # normal case
	    if (($curr_item + 1) * $item_size < $#$CCD_list)
	    {
		&PS_insert_endif($page_block,
				       $scale_factor,
				       $$CCD_list[$curr_item * $item_size + 2],
				       $$CCD_list[($curr_item + 1) * $item_size
						 + 2] 
				       - $$CCD_list[$curr_item * $item_size 
						   + 2] - 1,
				       $has_no_else,
				       $$CCD_list[$curr_item * $item_size + 6],
				       $$CCD_list[$curr_item * $item_size + 4]
				       );
	    }

	    # last endif in file
	    else
	    {
		&PS_insert_endif($page_block,
				       $scale_factor,
				       $$CCD_list[$curr_item * $item_size + 2],
				       $file_loc
				       - $$CCD_list[$curr_item * $item_size 
						   + 2],
				       $has_no_else,
				       $$CCD_list[$curr_item * $item_size + 6],
				       $$CCD_list[$curr_item * $item_size + 4]
				       );
	    }

	    # draw vertical dotted line
	    if ($has_no_else)
	    {
		# find the matching node
		for ($j = $curr_item - 1; 
		     ($j >= 0) 
		     && ($$CCD_list[$j * $item_size + 5]
			 != $$CCD_list[$curr_item * $item_size + 5]
			 );
		     $j--)
		{
		}
		&PS_draw_vert_line($page_block,
					 $scale_factor,
					 $$CCD_list[$curr_item * $item_size
						    + 4]
					 - $$CCD_list[$j * $item_size + 4] - 1,
					 1,
					 $$CCD_list[$j * $item_size + 6],
					 $$CCD_list[$j * $item_size + 4] + 1
					 );
	    }

	    $curr_x = $$CCD_list[$curr_item * $item_size + 6] + 1;
	    $curr_y = $$CCD_list[$curr_item * $item_size + 4] + 2;
	}
	$curr_item++;
    } # matches outer while loop

    return ($curr_item, $curr_x, $curr_y);
}

1;
