#!/usr/local/bin/perl -w
#
# $Source: /tmp_mnt/net/dworshak2/users/student/evans911/prog/cs481/src/RCS/process_c_file.pl,v $
# $Author: evans911 $
# Current revision: $Revision: 0.26 $
# Last modified: $Date: 1996/12/05 18:14:28 $
# State of current revision: $State: Exp $
# Currently locked by: $Locker:  $
#
# Description: This set of subroutines reads a C file and pulls out lines
#              with conditional compilation directives.  Quoted strings, 
#              C style comments, and C++ style comments are all handled
#              properly.
#              
# $Log: process_c_file.pl,v $
# Revision 0.26  1996/12/05 18:14:28  evans911
# Fixed problem with program going into infinite loop on non-C files.
# The problem had to do with not always verifying that reads from the
# input file were successful.
#
# Revision 0.25  1996/11/13 00:07:23  evans911
# Cleaned up comments.
#
# Revision 0.24  1996/11/08 00:31:03  evans911
# Added fields to @output to pad the stack, so that assign_postion()
# doesn't have to.
#
# Revision 0.23  1996/11/07 20:51:56  evans911
# Added proper handling of '...' strings, so that '/*' doesn't
# cause a problem any more.
#
# Revision 0.22  1996/10/29 23:59:32  evans911
# Removed debugging spew.
#
# Revision 0.21  1996/10/29 22:52:51  evans911
# Duh.  ...$ ) is NOT the same as ...$) in a regexp.  I changed ...$ )
# back to ...$) in two places.  Backslash continuation has been
# broken for several revisions, so if any older versions are used,
# make sure to fix this problem.
#
# Revision 0.20  1996/10/28 22:29:13  evans911
# Modified output array so that there are now three less values
# being output for each #CCD
#
# Revision 0.19  1996/10/24 16:27:34  evans911
# Added comments.
#
# Revision 0.18  1996/10/10 06:13:14  evans911
# Added code in scan_quote() to handle \\.  Also commented out unnecessary
# parts of process_file().  This version has been thoroughly tested
# against about 5.5 million lines of code plus the regression suite.
# At this point in time, this code correctly handles everything I can
# throw at it, so it may be free of defects.  (Yeah, right. =) )
#
# Revision 0.17  1996/10/09 22:09:32  evans911
# Completely rewrote scan_comment to handle \" and \ continuation.
#
# Revision 0.16  1996/10/09 02:07:37  evans911
# Fixed a number of problems with '"', "", \ (with space after it), etc.
# This version still works correctly on all of the verified tests, but
# fails on squeeze.c.  I'm not sure yet, but this could be from
# incorrectly handling \" inside of a quoted string.  I'll figure this
# one out later.
#
# Revision 0.15  1996/10/02 01:38:22  evans911
# Now strips white space from the beginning and end of expressions
# so that the summary report can properly reduce " X " and "X" to the
# same expression.
#
# Revision 0.14  1996/10/02 00:41:09  evans911
# Commented out debugging spew.
#
# Revision 0.13  1996/10/02 00:32:04  evans911
# Added #elif to V(g) and V`(g).
# Added optional white space before and after the # in a #CCD.
#
# Revision 0.12  1996/09/30 06:29:25  evans911
# Added code to handle backslash continuation.  It appears to work, but
# I can't be sure because I don't know what the regression testing is
# supposed to output.
#
# Revision 0.11  1996/09/18 20:15:21  evans911
# Fixed defect that caused an undef to be returned instead of #else
# or #elif.
#
# Revision 0.10  1996/09/18 02:03:47  evans911
# Minor fixes that make $. valid, and added call to &vgprime().
#
# Revision 0.9  1996/09/17 23:25:51  evans911
# Should work now.  Code V(g) has been removed.
#
# Revision 0.7  1996/09/07 07:32:20  evans911
# Commented out the test driver.
#
# Revision 0.6  1996/09/07 07:31:23  evans911
# Added code to handle C++ style comments.  In the process, I rewrote most
# of the logic for establishing precedence between /*, //, and ".  It's
# much better now.
#
# Revision 0.5  1996/09/07 06:34:04  evans911
# Added #elif and #else directives to search list.
# Added -w flag to first line.
#
# Revision 0.4  1996/09/06 19:31:16  evans911
# Fixed a minor bug.  If two comments began on the same line, text got lost.
# Added error printout if the end of file was reached while still in a
# comment or quoted string.
#
# Revision 0.3  1996/09/06 18:06:37  evans911
# Commented out all of the debugging printout and the test driver at
# the top of the file.
#
# Revision 0.2  1996/09/06 04:30:35  evans911
# It works!  There are lots of debugging prints that will need to be
# taken out, but the program passes every test we've thought of so far.
# It correctly strips out comments and quotes, and recognizes conditional
# compilation directives and their associated expressions.
# (Jason and Sean).
#
# Revision 0.1  1996/09/04 00:59:49  evans911
# Initial revision.  No functionality.
#

#
# Subroutine: scan_comment(*IN, $Line, $filename)
# Arguments: *IN : file descriptor for file being parsed
#            $Line : contains remainder of current line
#            $filename : name of file
#            
# Return Values: $Rest : remainder of line after end of comment
#                
# Globals Used:
#               
# Required Files: 
#                 
# Author: Sean Jones
#
# Description: This function accepts a line in which a comment block 
#              begins, and returns any and all code on the same line 
#              as and following the end of the comment block.
#              
sub scan_comment
{
    my($Line, $filename);
    (*IN, $Line, $filename) = @_;
    my $Comment;
    my $Rest;

    # Read to end of comment
    while (($Line !~ /\*\//s)&&($Line)){
	$Line = <IN>;
	if (!defined($Line))
	{
	    print STDERR "Error in $filename: reached end of file"
		. " without reaching end of comment.\n";
	}
    }

    # Return everything after the end of the comment
    ($Comment,$Rest) = split(/\*\//,$Line, 2); # split on "*/"
    
    return ($Rest);
}

#
# Subroutine: scan_tick(*IN, $line, $filename)
# Arguments: *IN : file descriptor for file being parsed
#            $line : contains remainder of current line
#            $filename : name of file
#            
# Return Values: $quote : quoted text, minus the first quote
#                $rest : everything after the quoted text
#                
# Globals Used: *IN : file handle to get input from
#               
# Required Files: 
#                 
# Author: Jason Evans
#
# Description: This function accepts a line following a found ' operator.
#              It finds the matching end ' and returns both the entire
#              quote (including the end ') and any and all code following
#              the end ' on the same line.
#
sub scan_tick
{
    my ($line, $filename);
    (*IN, $line, $filename) = @_;
    my $quote;
    my $rest;
    my ($end_of_quote) = 0;
    my ($before_separator, $separator);

    # Initialize var to store quote
    $quote = "";

    # loop until end of quote
    until ($end_of_quote)
    {
	# look for (\\), (\'), ('), or (\)
	if ($line =~ /(\\\\|\\\'|\'|\\\s*$)/)
	{
	    $separator = $1;
	    if ($separator =~ /\\\\/)
	    {
		($before_separator, $line) = split(/\\\\/, $line, 2);
		$quote = $quote . $before_separator . $separator;
	    }
	    elsif ($separator =~ /\\\'/) # matched (\')
	    {
		($before_separator, $line) = split(/\\\'/, $line, 2);
		$quote = $quote . $before_separator . $separator;
	    }
	    elsif ($separator =~ /\'/) # matched (')
	    {
		$end_of_quote = 1;
		($before_separator, $rest) = split(/\'/, $line, 2);
		$quote = $quote . $before_separator . $separator;
	    }
	    else #elsif ($separator =~ /\\\s*$/) # matched (\)
	    {
		chop $line;
		$quote = $quote . $line;
		$line = <IN>;
		if (!defined($line))
		{
		    print STDERR "Error in $filename: reached end of file"
			. " without reaching end of single quoted string.\n";
		    exit(-1);
		}
	    }
	}
        else
	{
	    $quote = $quote . $line;
	    $line = <IN>;
	    if (!defined($line))
	    {
		print STDERR "Error in $filename: reached end of file"
		    . " without reaching end of single quoted string.\n";
		exit(-1);
	    }
	}
    }

    return ($quote, $rest);
}

#
# Subroutine: scan_quote(*IN, $line, $filename)
# Arguments: *IN : file descriptor for file being parsed
#            $line : contains remainder of current line
#            $filename : name of file
#            
# Return Values: $quote : quoted text, minus the first quote
#                $rest : everything after the quoted text
#                
# Globals Used: *IN : file handle to get input from
#               
# Required Files: 
#                 
# Author: Jason Evans
#
# Description: This function accepts a line following a found " operator.
#              It finds the matching end " and returns both the entire
#              quote (including the end ") and any and all code following
#              the end " on the same line.
#
sub scan_quote
{
    my ($line, $filename);
    (*IN, $line, $filename) = @_;
    my $quote;
    my $rest;
    my ($end_of_quote) = 0;
    my ($before_separator, $separator);

    # Initialize var to store quote
    $quote = "";

    # loop until end of quote
    until ($end_of_quote)
    {
	# look for (\\), (\"), ("), or (\)
	if ($line =~ /(\\\\|\\\"|\"|\\\s*$)/)
	{
	    $separator = $1;
	    if ($separator =~ /\\\\/)
	    {
		($before_separator, $line) = split(/\\\\/, $line, 2);
		$quote = $quote . $before_separator . $separator;
	    }
	    elsif ($separator =~ /\\\"/) # matched (\")
	    {
		($before_separator, $line) = split(/\\\"/, $line, 2);
		$quote = $quote . $before_separator . $separator;
	    }
	    elsif ($separator =~ /\"/) # matched (")
	    {
		$end_of_quote = 1;
		($before_separator, $rest) = split(/\"/, $line, 2);
		$quote = $quote . $before_separator . $separator;
	    }
	    else #elsif ($separator =~ /\\\s*$/) # matched (\)
	    {
		chop $line;
		$quote = $quote . $line;
		$line = <IN>;
		if (!defined($line))
		{
		    print STDERR "Error in $filename: reached end of file"
			. " without reaching end of double quoted string.\n";
		    exit(-1);
		}
	    }
	}
        else
	{
	    $quote = $quote . $line;
	    $line = <IN>;
	    if (!defined($line))
	    {
		print STDERR "Error in $filename: reached end of file"
		    . " without reaching end of double quoted string.\n";
		exit(-1);
	    }
	}
    }

    return ($quote, $rest);
}

#
# Subroutine: process_file($infile, $outfile)
# Arguments: $infile : name of file to process
#            
# Return Values: $LOC : lines of code in file
#                $curr_CCD_Vg : V(g) of #CCD logic for entire file
#                $curr_CCD_Vgp : V'(g) of #CCD logic for entire file
#                @output : array of directives and their associated info:
#                   $directive : #CCD                                    
#                   $expression : expression for #CCD                    
#                   $line_num_for_directive : line #CCD starts on        
#                   top line length for #CCD (set to 1)
#                   physical y position for #CCD (set to -1)
#                   logical x position for #CCD (set to -1)
#                   physical x position for #CCD (set to 0)
#                
# Globals Used: *INFILE : filehandle for file to be parsed.
#               
# Required Files: 
#                 
# Author: Jason Evans
#
# Description: Given a filename, parse the file and find all preprocessor
#              conditionals (#if, #ifdef, #ifndef, #elif, #els, #endif).  
#              Also handles comments and quoted strings with the help of 
#              scan_comment() and scan_quote().  C++ style comments are
#              handled directly in this function.
#
sub process_file
{
    my($infile) = @_;
    my($has_directive) = 0;
    my($line_num_for_directive);
    my($line_text, $curr_text, $processed_line, $processed_wo_quotes, 
       $inside_quote);
    my($eol) = 0;
    my($pre_separator, $post_separator, $separator);
    my(@output);
    my($directive, $expression);
    my($curr_CCD_Vg) = 1;
    my($curr_CCD_Vgp) = 1;
    
    # open file
    (open(INFILE, "$infile")) ||
	die "process_file(): can't open \"$infile\".";

    # turn buffering off
    select(INFILE);
    $| = 1;
    select(STDOUT);
    
    #
    # Scan input file, and save info about lines with directives.
    #
    while ($line_text = <INFILE>)
    {
	# figure out if line starts with a preprocessor conditional
	$processed_line = "";
	$processed_wo_quotes = "";
	$has_directive = 0;
	if ($line_text =~ (/^\s*\#\s*if[^A-Za-z]?/)
	    || ($line_text =~ /^\s*\#\s*else[^A-Za-z]?/)
	    || ($line_text =~ /^\s*\#\s*elif[^A-Za-z]?/)
	    || ($line_text =~ /^\s*\#\s*ifdef[^A-Za-z]?/)
	    || ($line_text =~ /^\s*\#\s*ifndef[^A-Za-z]?/)
	    || ($line_text =~ /^\s*\#\s*endif[^A-Za-z]?/))
	{
	    $has_directive = 1;
	    $line_num_for_directive = $.;
	}

	#
	# Keep reading in text until the "logical" end of a line is reached.
	# This means continuing on \ continuation or quoted strings with
	# embedded carriage returns.
	#
	$eol = 0;
	$curr_text = $line_text;
	until ($eol)
	{
	    #
	    # Check for comments and strings ('...', "...", /*...*/, //...),
	    # as well as \ continuation.
	    #
	    if ($curr_text =~ /(\'|\"|\/\/|\/\*|\\\s*$)/)
	    {
		$separator = $1;
		if ($separator =~ /^\'/) # found a quote (')
		{
		    ($pre_separator, $post_separator) = split(/\'/, 
							      $curr_text, 
							      2);

		    $processed_wo_quotes = $processed_line . $pre_separator;
		    $processed_line = $processed_line . $pre_separator . "\'";

		    ($inside_quote, $curr_text) = &scan_tick(*INFILE,
							      $post_separator,
							      $infile);

		    $processed_line = $processed_line . $inside_quote;
		}
		elsif ($separator =~ /^\"/) # found a quote (")
		{
		    ($pre_separator, $post_separator) = split(/\"/, 
							      $curr_text, 
							      2);

		    $processed_wo_quotes = $processed_line . $pre_separator;
		    $processed_line = $processed_line . $pre_separator . "\"";

		    ($inside_quote, $curr_text) = &scan_quote(*INFILE, 
							      $post_separator,
							      $infile);

		    $processed_line = $processed_line . $inside_quote;
		}
		elsif ($separator =~ /\/\//) # found a comment (//)
		{
		    $eol = 1;

		    ($pre_separator, $post_separator) = split(/\/\//, 
							      $curr_text, 
							      2);
		    $processed_line = $processed_line . $pre_separator;
		    $processed_wo_quotes = $processed_wo_quotes 
			. $pre_separator;
		}
		elsif ($separator =~ /\/\*/) # found a comment (/*)
		{
		    ($pre_separator, $post_separator) = split(/\/\*/, 
							      $curr_text, 
							      2);
		    $processed_line = $processed_line . $pre_separator;
		    $processed_wo_quotes = $processed_wo_quotes 
			. $pre_separator;

		    ($curr_text) = &scan_comment(*INFILE,
						 $post_separator,
						 $infile);
		}
		else # ($separator =~ /\\/) # found a \ continuation (\)
		{
		    $curr_text =~ /(.*)\\\s*$/;
		    ($processed_line) = $processed_line . $1;
		    ($processed_wo_quotes) = $processed_wo_quotes . $1;
		    $curr_text = <INFILE>;
		    if (!defined($curr_text))
		    {
			print STDERR "Error in $infile: reached end of file"
			    . " without reaching end of backslash continued"
				. " conditional directive.\n";
			exit(-1);
		    }
		}
	    }
	    else
	    {
		$eol = 1;
		if ($curr_text =~ /.*\n$/s)
		{
		    chop($curr_text);
		}
		$processed_line = $processed_line . $curr_text;
		$processed_wo_quotes = $processed_wo_quotes . $curr_text;
	    }
	}

	#
	# If the current line is one of the directives we are looking for,
	# calculate its cumulative V(g) and V'(g), and output all the
	# important information to an array that will be returned once
	# the whole file has been processed.
	#
	if ($has_directive)
	{
	    # calculate #CCD V(g) for line and add to $curr_CCD_Vg
	    if ($processed_wo_quotes =~ /^\s*\#\s*(elif|if)/)
	    {
		$curr_CCD_Vg += 1;
		$curr_CCD_Vgp += 1;
	    }

	    #
	    # Strip out white space before and after the #, and break the
	    # line up into the directive and the expression.
	    #
	    $processed_line =~ 
		/^\s*\#\s*(ifndef|ifdef|if|endif|else|elif)([^\0]*)/;
	    $directive = "\#" . $1;
	    $expression = $2;

	    #
	    # Make sure that expression is defined before calling vgprime()
	    # or outputting it to @output.
	    #
	    unless (defined($expression))
	    {
		$expression = "";
	    }

	    ($curr_CCD_Vgp) += &vgprime($expression);

	    # strip white space from beginning and end of expression
	    if ($expression =~ /^\s*(\S.*\S|\S)\s*$/)
	    {
		$expression = $1;
	    }

	    # output information to array
	    (@output) = (@output,
			 $directive,
			 $expression,
			 $line_num_for_directive,
			 1,
			 -1,
			 -1,
			 0
			 );
	}
    }

    $LOC = $.;

    close(INFILE);

    return ($LOC, $curr_CCD_Vg, $curr_CCD_Vgp, @output);
}

1;
