#!/usr/local/bin/perl
#
# $Source: /tmp_mnt/net/dworshak2/users/student/evans911/prog/cs481/src/RCS/gen_report.pl,v $
# $Author: jone9333 $
# Current revision: $Revision: 0.13 $
# Last modified: $Date: 1996/11/08 01:02:48 $
# State of current revision: $State: Exp $
# Currently locked by: $Locker:  $
#
# Description: Contains subroutine for printing a summary report.
#              
# $Log: gen_report.pl,v $
# Revision 0.13  1996/11/08 01:02:48  jone9333
# Made the stack parameter call by reference.
#
# Revision 0.12  1996/11/08 00:12:09  jone9333
# Made sublist size a parameter.
#
# Revision 0.11  1996/10/29 01:09:59  evans911
# Changed hard coded $size to 3.
#
# Revision 0.10  1996/10/28 22:19:34  evans911
# Removed white space at end of file.
# Added file comment.
#
# Revision 0.9  1996/10/23 22:53:51  jone9333
# Reduced indentation in summary report.
# Right justified all numerical output.
#
# Revision 0.8  1996/10/17 20:34:14  jone9333
# Added full commenting.
# Changed directive summary to print totals for all 6 possible directives.
#
# Revision 0.7  1996/10/09 22:38:12  evans911
# Added comment block.
#
# Revision 0.6  1996/09/25 03:44:51  jone9333
# Changed parameters to match changes in cca.pl.
# Can now selectively create directive/expression summaries and/or
# display a complete list of conditional logic.
#
# Revision 0.5  1996/09/23 21:39:46  jone9333
# Added counters to the summary report to display number of instances
# of each directive / expression.
# This was done with perl hash data structures.
#
# Revision 0.4  1996/09/19 01:37:03  jone9333
# Added code to generate a list of expressions used in the source
# code file.
#
# Revision 0.3  1996/09/18 02:06:23  evans911
# Fixed syntax errors and made other small fixes.
#
# Revision 0.2  1996/09/18 01:40:56  jone9333
# Now reports total LOC, V(g), V'(g).
#
# Revision 0.1  1996/09/18 01:26:46  evans911
# Initial revision.
#
#

#
# Subroutine: gen_report($C, $D, $E, $L, $T, $V, $outfile, $infile,
#                        $LOC, $CCD_Vg, $CCD_Vgp, \@CCD_list);
#
# Arguments: $C :        print comprehensive list switch
#            $D :        print directive summary switch
#            $E :        print expression summary switch
#            $L :        print total lines of code switch
#            $T :        print total lines of conditional compilation 
#                        logic code switch
#            $V :        print V(g) and V'(g) switch
#            $outfile :  output file name
#            $infile :   input file name
#            $LOC :      total lines of code
#            $CCD_Vg :   total V(g) 
#            $CCD_Vgp :  total V'(g)
#            $size :     number of elements in a sublist within @CCD_list
#            $CCD_list : reference to complete list of all conditional 
#                        compilation code
#            
#            
# Return Values: 
#                
# Globals Used: 
#
# Required Files: 
#                 
# Author: Sean Jones
#
# Description: This function checks the command line switches which
#              have been set at execution, and prints the appropriate
#              analysis results.
#              
sub gen_report
{
    my $num_items;
    my $list_index;
    my %ccds;
    my %exprs;
    my $CCD_LOC;
    my($C, $D, $E, $L, $T, $V, $outfile, $infile, 
       $LOC, $CCD_Vg, $CCD_Vgp, $size, $CCD_list) = @_;
    
    # Set output channel
    if (defined($outfile)){
	open(OUT,">$outfile")||
	    die"Couldn't open $outfile.\n";
	select(OUT);
    }

    printf("\nSummary Report for %s :\n\n",$infile);

    # Display command line switches specified
    printf("Analysis Options Specified:\n\n");
    if($C){
	printf("C : Display Comprehensive Conditional Logic Listing\n");
    }
    if($D){
	printf("D : Display Conditional Directive Summary\n");
    }
    if($E){
	printf("E : Display Conditional Expression Summary\n");
    }
    if($L){
	printf("L : Display Total Lines of Code\n");
    }
    if($T){
	printf("T : Display Total Lines of Conditional Compilation Logic\n");
    }
    if($V){
	printf("V : Display V(g) and V'(g) for Conditional Logic\n");
    }

    printf("\nCode Analysis Summary:\n\n");
    if ($L){
	printf("\tTotal Lines of Code:\t\t%6d\n",$LOC);
    }
    if ($T){
	$CCD_LOC = ($#$CCD_list + 1) / $size;
	printf("\tTotal Lines of CC Logic:\t%6d\n",$CCD_LOC);
    }
    if ($V){
	printf("\tTotal V(g) for CCDs:\t\t%6d\n",$CCD_Vg);
	printf("\tTotal V'(g) for CCDs:\t\t%6d\n",$CCD_Vgp);
    }

    if ($D){
	printf("\nDirective Use Summary:\n\n");
	$num_items = size_stack($size,@$CCD_list);
	
	#initialize hash variable %ccds to all 0's
	$ccds{"#if"} = 0;
	$ccds{"#ifdef"} = 0;
	$ccds{"#ifndef"} = 0;
	$ccds{"#else"} = 0;
	$ccds{"#elif"} = 0;
	$ccds{"#endif"} = 0;

	#for each directive found, increment appropriate counter
	for($list_index = 0;$list_index < $num_items; $list_index++){
	    $_ = $$CCD_list[$list_index * $size + 0];
	    if ((/\w+/)&&(defined($_))){
		$ccds{$_} += 1;
	    }
	}

	#output summary data
	printf("\t%5d\t#if\n",$ccds{"#if"});
	printf("\t%5d\t#ifdef\n",$ccds{"#ifdef"});
	printf("\t%5d\t#ifndef\n",$ccds{"#ifndef"});
	printf("\t%5d\t#else\n",$ccds{"#else"});
	printf("\t%5d\t#elif\n",$ccds{"#elif"});
	printf("\t%5d\t#endif\n",$ccds{"#endif"});
    }	

    if ($E){
	#Print list of expressions used
	printf("\nExpression Use Summary:\n\n");
	$num_items = size_stack($size,@$CCD_list);

	#for each expression, increment corresponding counter
	for($list_index = 0;$list_index < $num_items; $list_index++){
	    $_ = $$CCD_list[$list_index * $size + 1];
	    if ((/\w+/)&&(defined($_))){
		if (!defined($exprs{$_})){
		    $exprs{$_} = 0;
		}
		$exprs{$_} += 1;		
	    }
	}
	
	#sort expressions
	@keys = keys(%exprs);
	@keys = sort(@keys);

	#print summary
	foreach (@keys){
	    printf("\t%5d\t%s\n",$exprs{$_},$_);
	}	    
    }

    if ($C){
	#Print Comprehensive Listing
	printf("\nComplete Conditional Logic Listing:\n\n");
	printf("Line#\tCCD\t\tExpression\n");
	printf( "----------------------------------------------\n");

	$num_items = size_stack($size,@$CCD_list);
	
	for($list_index = 0;$list_index < $num_items; $list_index++){
	    $lin = $$CCD_list[$list_index * $size + 2];
	    $_ = $$CCD_list[$list_index * $size + 0];
	    if ((/\w+/)&&(defined($_))){
		printf("%5d\t%s\t\t",$lin,$_);
	    }
	    $_ = $$CCD_list[$list_index * $size + 1];
	    if ((/\w+/)&&(defined($_))){
		printf("%s\n",$_);
	    }
	    else{
		printf("\n");
	    }
	}
    }

    printf("\n*** End of Analysis Summary ***\n");

    close(OUT);
}

1;
