#!/usr/local/bin/perl -w
#
# $Source: /tmp_mnt/net/dworshak2/users/student/evans911/prog/cs481/src/RCS/ps.pl,v $
# $Author: evans911 $
# Current revision: $Revision: 1.18 $
# Last modified: $Date: 1996/11/21 22:24:29 $
# State of current revision: $State: Exp $
# Currently locked by: $Locker:  $
#
# Description: This set of routines is used to build a PostScript graph that
#              represents the conditional branching of C code.  For the most
#              part, the individual routines are self-explanatory, but a
#              couple of concepts concerning the PostScript code aren't
#              immediately evident.
#              
#              The PostScript functions treat the drawing space as though
#              it were tiled into squares.  All symbol drawing is done in
#              proportion to the size of the squares.
#              
#              Scaling of the graph is achieved by scaling the coordinate
#              axes.  This means that everything gets drawn bigger or smaller,
#              depending on the scale factor.  Line width even varies.
#              
# $Log: ps.pl,v $
# Revision 1.18  1996/11/21 22:24:29  evans911
# Removed commented out PS code.
#
# Revision 1.17  1996/11/21 22:14:01  evans911
# Fixed the filename/page coordinate size problem.
#
# Revision 1.16  1996/11/13 02:15:12  evans911
# Reversed the if checks in all of the PS function insertion routines
# so that fewer conditions have to be checked.
#
# Revision 1.15  1996/11/13 01:57:19  evans911
# Updated comment on line 123.
#
# Revision 1.14  1996/11/13 01:27:49  evans911
# Updated comments.
#
# Revision 1.13  1996/11/11 19:13:44  evans911
# Made modifications to add two more scale factors.
#
# Revision 1.12  1996/11/07 22:21:33  evans911
# Fixed file name on printout.
# Made modifications that make sure <> nodes get printed on the
# proper pages.
#
# Revision 1.11  1996/11/07 20:06:32  evans911
# Modified functions to use an associative array of pages and insert
# a function call on each page that a node or line is visible.  This
# still has some problems.  For example, expressions may get choped
# off at a page boundary.
#
# Revision 1.10  1996/10/25 17:54:46  evans911
# Modified PS_draw_top_line() to work with optional LOC text.
#
# Revision 1.9  1996/10/24 16:21:19  evans911
# Added comments.
#
# Revision 1.8  1996/10/23 21:15:45  evans911
# Removed printing of BLOC.
#
# Revision 1.7  1996/10/17 17:00:32  evans911
# Duh, I think the \ protection of expressions may be working now.
# I was doing the substitutions in the wrong order.
#
# Revision 1.6  1996/10/17 15:13:01  evans911
# Fixed character substitution code in PS_insert_if().
#
# Revision 1.5  1996/10/16 23:51:23  evans911
# Added \ protection to [()#] for expressions.
#
# Revision 1.4  1996/10/16 01:24:39  evans911
# This version may be finished, except that BLOC formatting may need
# changed.
#

#
# Subroutine: PS_create_page_block($filename, $xpages, $ypages)
# Arguments: $filename : name of input code file
#            $xpages   : # of pages in the horizontal direction
#            $ypages   : # of pages in the vertical direction
#            
# Return Values: %page_block : associative array of pages, keyed by (x, y)
#                
# Globals Used: 
#               
# Required Files: 
#                 
# Author: Jason Evans
#
# Description: Given the input code filename and the number of pages in the
#              x and y directions, create an associative array of blank 
#              pages.
#              
sub PS_create_page_block
{
    my ($filename, $xpages, $ypages) = @_;
    my (%page_block) = "";
    my ($i, $j, $page);
    my ($page_template) = "%%Page: %PAGE% %PAGE%
%%BeginPageSetup
%%EndPageSetup
gsave
  (%FILENAME% \(%XPAGE%, %YPAGE%\)) PageNum
  gsave
    %XPAGE% %YPAGE% XYPage
  grestore
  showpage
grestore
%%PageTrailer

";

    # create pages
    $page = 1;
    for ($i = 0; $i < $ypages; $i++)
    {
	for ($j = 0; $j < $xpages; $j++)
	{
	    $_ = $page_template;
	    $_ =~ s/\%PAGE\%/$page/g;
	    $_ =~ s/\%XPAGE\%/$j/g;
	    $_ =~ s/\%YPAGE\%/$i/g;
	    $_ =~ s/\%FILENAME\%/$filename/g;
	    $page_block{$j, $i} = $_;
	    $page++;
	}
    }
    
    # write filename instead of (x,y) page coordinate for 0,0
#    $page_block{0, 0} =~ s/\(0, 0\) PageNum/\($filename\) PageNum/g;

    return %page_block;
}

#
# Subroutine: PS_create_file($PSfilename, $scale_factor, $xpages, $ypages, 
#                            $page_block)
# Arguments: $PSfilename   : name of graph output file
#            $scale_factor : scale factor for graph
#            $xpages       : x dimension of image in pages
#            $ypages       : y dimension of image in pages
#            
# Return Values: 
#                
# Globals Used: 
#               
# Required Files: 
#                 
# Author: Jason Evans
#
# Description: Outputs page blocks for all valid (x,y) coordinates, outputs
#              the necessary closing PS-DSC gunk, then closes the file handle.
#              
sub PS_create_file
{
    my ($PSfilename, $scale_factor, $xpages, $ypages, $page_block) = @_;
    my ($i, $j);
    my ($before_pages) = "%!PS-Adobe-3.0
%%EndComments
%%BeginProlog
%------------- Procedure Definitions -------------

% sets font to bold
% float(point size) BF
/BF
{
  dup 
  /fontsize exch def

  /Helvetica-Bold findfont 
  exch scalefont % uses stack operand
  setfont
}
def

% draws a decision diamond for #if, #ifdef, and #ifndef
%
% str(#CCD)
% str(expression)
% int(line num)
% int(top loc)
% int(bottom loc)
% int(dotted top)
% int(x coord)
% int(y coord)
% DrawIf
/DrawIf
{
  gsave
    newpath
    square neg exch square exch translate % move to upper left corner of square

    % draw diamond and lines
    gsave
      0.5 square -0.25 square moveto
      0.25 square -0.25 square rlineto
      -0.25 square -0.25 square rlineto
      -0.25 square 0.25 square rlineto
      closepath
      stroke
    grestore

    gsave
      1 eq
      {
	[1 1] 1 setdash	
      }
      if

      0.5 square 0 square moveto
      0 square -0.25 square rlineto
      stroke
    grestore

    gsave
      0.75 square -0.5 square moveto
      0.25 square 0 square rlineto
      stroke
    grestore

    gsave
      [1 1] 0 setdash
      0.5 square -0.75 square moveto
      0 square -0.25 square rlineto
      stroke
    grestore

    % write lower LOC
    gsave
      0.53 square -0.82 square moveto
      show
    grestore

    % write upper LOC
    gsave
      0.53 square -0.18 square moveto
      show
    grestore

    % write line number
    gsave
      0.21 square -0.525 square moveto
      dup stringwidth pop neg 0 rmoveto show
    grestore

    %write expression
    gsave
      0.78 square -0.42 square moveto
      show
    grestore

    % write #if in center of diamond
    gsave
      0.5 square -0.5 square moveto
      dup stringwidth pop neg 2 div -0.025 square rmoveto
      show
    grestore
  grestore
}
bind def

% draws a triangle for #else
%
% str(line num)
% int(x coord)
% int(y coord)
% DrawElse
/DrawElse
{
  gsave
    newpath
    square neg exch square exch translate % move to upper left corner of square

    % draw triangle and lines
    gsave 
      0.375 square -0.125 square moveto
      0 square -0.5 square rlineto
      0.5 square 0 square rlineto
      closepath
      stroke
    grestore

    gsave
      0.5 square 0 square moveto
      [1 1] 1 setdash
      0 square -0.25 square rlineto
      stroke
    grestore

    gsave
      0.75 square -0.5 square moveto
      0.25 square 0 square rlineto

      stroke
    grestore

    % write line number
    gsave
      0.21 square -0.525 square moveto
      dup stringwidth pop neg 0 rmoveto show
    grestore

    % write #else in center of diamond
    gsave
      (\\#else) 
      0.435 square -0.525 square moveto
      show
    grestore
  grestore
} 
bind def

% draws a triangle for #endif
%
% str(line num)
% str(LOC)
% int(line flag) 1 if the line is to be drawn
% int(x coord)
% int(y coord)
% DrawEndif
/DrawEndif
{
  gsave
    newpath
    square neg exch square exch translate % move to upper left corner of square

    % draw circle and lines
    gsave 
      0.5 square -0.5 square 0.18 square 0 360 arc
      stroke
    grestore

    gsave
      1 eq
      {
	0.5 square 0 square moveto
	[1 1] 1 setdash
	0 square -0.32 square rlineto
	stroke
      }
      if
    grestore

    gsave
      0.68 square -0.5 square moveto
      0.32 square 0 square rlineto
      stroke
    grestore

    gsave
      0.5 square -0.68 square moveto
      0 square -0.32 square rlineto
      stroke
    grestore

    % write lower LOC
    gsave
      0.53 square -0.82 square moveto
      show
    grestore

    % write line number
    gsave
      0.21 square -0.525 square moveto
      dup stringwidth pop neg 0 rmoveto show
    grestore

    % write #endif in center of diamond
    gsave
      0.35 square -0.525 square moveto
      (\\#endif)
      show
    grestore

  grestore
} 
bind def

% Draws a line from left to right, plus the corner and LOC
%
% string(LOC)
% int(length)
% int(x start)
% int(y start)
% DrawTopLine
/DrawTopLine
{
  gsave
    square neg exch square exch translate % move to upper left corner of square
    dup % copy length for later use
    gsave
      0 square -0.5 square moveto
      0.5 sub square 0 square rlineto
      0 square -0.5 square rlineto
      stroke
    grestore

    gsave
      0.47 sub square -0.65 square moveto
      show
    grestore

  grestore
}
bind def

% Draws a line from right to left, plus the corner
%
% int(length)
% int(x start)
% int(y start)
% DrawBottomLine
/DrawBottomLine
{
  gsave
    square neg exch square exch translate % move to upper left corner of square
    gsave
      0.5 square 0 square moveto
      0 square -0.5 square rlineto
      0.5 sub neg square 0 square rlineto
      stroke
    grestore
  grestore
}
bind def

% Draws a line from top to bottom
%
% int(length)
% int(is dotted)
% int(x start)
% int(y start)
% DrawVertLine
/DrawVertLine
{
  gsave
    square neg exch square exch translate % move to upper left corner of square
    gsave
      1 eq
      {
	[1 1] 1 setdash
      }
      if
      0.5 square 0 square moveto
      square neg 0 square exch rlineto
      stroke
    grestore
  grestore
}
bind def

%
% miscellaneous defs
%
/XPt {576} bind def
/YPt {720} bind def
/square {36 mul} bind def

/ScaleFactor {%SCALE%} bind def

/Scale {ScaleFactor div} bind def

/XPage {XPt Scale mul} bind def
/YPage {YPt Scale mul} bind def

/XYPage 
{
  % set clipping path for this page
  newpath
  -1 square 0 moveto
  XPt Scale 0 rlineto
  0 YPt neg Scale rlineto
  XPt neg Scale 0 rlineto
  closepath clip
  0.7 setgray
  clippath stroke
  0 setgray

  % translate to this page
  YPage exch XPage neg exch translate
}
bind def

% prints page number on page
%
% str(page number)
% PageNum
/PageNum
{
  gsave
    1 ScaleFactor div dup scale
    /Helvetica-Bold findfont 15 scalefont setfont
    newpath
    -0.95 square ScaleFactor mul
    6
    moveto
    show
  grestore
}
bind def

% End of procedure definitions
%EndProlog

%%BeginSetup
/Helvetica-Bold findfont 0.1 square scalefont setfont
20 72 11 mul 36 sub translate % move origin to top left of graph
ScaleFactor ScaleFactor scale
1 square 0 translate
%%EndSetup
";

    my ($after_pages) = "%%Trailer
%%EOF
";

    open (PSOUT, ">$PSfilename")
	|| die "Fatal Error - Couldn't open $PSfilename for writing.\n";

    # subsitute the scale factor in place of %SCALE% in $HeaderData;
    $before_pages =~ s/\%SCALE\%/$scale_factor/g;

    # spew stuff to the file
    print PSOUT "$before_pages";

    for ($j = 0; $j < $ypages; $j++)
    {
      for ($i = 0; $i < $xpages; $i++)
      {
        print PSOUT "$$page_block{$i, $j}";
      }
    }

    print PSOUT "$after_pages";

    close PSOUT;
}

#
# Subroutine: PS_insert_if($page_block, $scale, $xpages, $ccd, $expression, 
#                          $line_num, $loc, $dotted_top, $xpos, $ypos);
#
# Arguments: $page_block : reference to associative array of page blocks
#            $scale      : scale factor for graph
#            $xpages     : # of pages in horizontal direction
#            $ccd        : Conditional compilation directive
#            $expression : Expression belonging to $ccd
#            $lin_num    : Line number $ccd appears on            
#            $loc        : Lines of code above $ccd
#            $dotted_top : If top line should be dotted, then it's == 1.
#            $xpos       : X position
#            $ypos       : Y position
#
# Return Values: 
#                
# Globals Used: 
#               
# Required Files: Jason Evans
#                 
# Description: Inserts PS code to draw a #if symbol.
#              
sub PS_insert_if
{
    my ($page_block, $scale, $xpages,
       $ccd, $expression, $line_num, $loc, $dotted_top, 
       $xpos, $ypos) = @_;
    my ($xpage_start, $xpage_end, $ypage, $i);
    my ($junk);

    # protect any special characters
    $expression =~ s/\\/\\\\/g;
    $expression =~ s/\)/\\\)/g;
    $expression =~ s/\(/\\\(/g;

    # figure out which pages to put it on
    if ($scale eq 'p') # scaled to one page
    {
	$xpage_start = 0;
	$xpage_end = 0;
	$ypage = 0;
    }
    else # must be 1, 2, 4, 8, or 16
    {
	$xpage_start = (($xpos) / (64 / $scale));
	($xpage_start, $junk) = split(/\./, $xpage_start, 2);

	$xpage_end = (($xpos + 1) / (64 / $scale));
	($xpage_end, $junk) = split(/\./, $xpage_end, 2);
	$xpage_end += 1;

	$ypage = (($ypos) / (80 / $scale));
	($ypage, $junk) = split(/\./, $ypage, 2);
    }


    for ($i = $xpage_start; ($i <= $xpage_end) && ($i < $xpages); $i++)
    {
	# insert call into page block
	$$page_block{$i, $ypage} 
	=~ s/$i $ypage XYPage/$i $ypage XYPage\n  \(\\$ccd\) \($expression\) \($line_num\) \($loc\) \(\) $dotted_top $xpos $ypos DrawIf/;
    }
}

#
# Subroutine: PS_insert_else($page_block, $scale, $line_num, $xpos, $ypos)
# Arguments: $page_block : reference to associative array of page blocks
#            $scale      : scale factor for graph
#            $line_num   : line number at which the #else occurs
#            $xpos       : x position
#            $ypos       : y position
#            
# Return Values: 
#                
# Globals Used: 
#               
# Required Files: 
#                 
# Author: Jason Evans
#
# Description: Inserts PS code to draw a #else symbol.
#              
sub PS_insert_else
{
    my ($page_block, $scale, 
     $line_num, $xpos, $ypos) = @_;
    my ($xpage_start, $xpage_end, $ypage, $i);
    my ($junk);

    # figure out which page to put it on
    if ($scale eq 'p') # scaled to one page
    {
	$xpage_start = 0;
	$xpage_end = 0;
	$ypage = 0;
    }
    else # must be 1, 2, 4, 8, or 16
    {
	$xpage_start = (($xpos) / (64 / $scale));
	($xpage_start, $junk) = split(/\./, $xpage_start, 2);

	$xpage_end = (($xpos + 1) / (64 / $scale));
	($xpage_end, $junk) = split(/\./, $xpage_end, 2);

	$ypage = (($ypos) / (80 / $scale));
	($ypage, $junk) = split(/\./, $ypage, 2);
    }


    # build list of pages to insert function call into
    for ($i = $xpage_start; $i <= $xpage_end; $i++)
    {
	# insert call into page block
	$$page_block{$i, $ypage} =~ s/$i $ypage XYPage/$i $ypage XYPage\n  \($line_num\) $xpos $ypos DrawElse/;
    }
}

#
# Subroutine: PS_insert_endif($page_block, $scale, $line_num, $loc, 
#                             $draw_above_line, $xpos, $ypos)
# Arguments: $page_block      : reference to associative array of page blocks
#            $scale           : scale factor for graph
#            $line_num        : Line number at which the #endif occurs
#            $loc             : LOC below the #endif
#            $draw_above_line : If top line should be drawn, it's == 1
#            $xpos            : X position
#            $ypos            : Y position
#            
# Return Values: 
#                
# Globals Used: 
#               
# Required Files: 
#                 
# Author: Jason Evans
#
# Description: Inserts PS code to draw a #endif symbol.
#              
sub PS_insert_endif
{
    my ($page_block, $scale, 
	$line_num, $loc, $draw_above_line, $xpos, $ypos) = @_;
    my ($xpage_start, $xpage_end, $ypage, $i);
    my ($junk);

    # figure out which page to put it on
    if ($scale eq 'p') # scaled to one page
    {
	$xpage_start = 0;
	$xpage_end = 0;
	$ypage = 0;
    }
    else # must be 1, 2, 4, 8, or 16
    {
	$xpage_start = (($xpos) / (64 / $scale));
	($xpage_start, $junk) = split(/\./, $xpage_start, 2);

	$xpage_end = (($xpos + 1) / (64 / $scale));
	($xpage_end, $junk) = split(/\./, $xpage_end, 2);

	$ypage = (($ypos) / (80 / $scale));
	($ypage, $junk) = split(/\./, $ypage, 2);
    }

    # build list of pages to insert function call into
    for ($i = $xpage_start; $i <= $xpage_end; $i++)
    {
	# insert call into page block
	$$page_block{$i, $ypage} 
	=~ s/$i $ypage XYPage/$i $ypage XYPage\n  \($line_num\) \($loc LOC\) $draw_above_line $xpos $ypos DrawEndif/;
    }
}

#
# Subroutine: PS_draw_top_line($page_block, $scale, $loc, $length, $xstart, 
#                              $ystart)
# Arguments: $page_block      : reference to associative array of page blocks
#            $scale           : scale factor for graph
#            $loc             : LOC for following block
#            $length          : x length
#            $xstart          : x starting position
#            $ystart          : y starting position
#            
# Return Values: 
#                
# Globals Used: 
#               
# Required Files: 
#                 
# Author: Jason Evans
#
# Description: Inserts PS code to draw a top line and LOC.
#              
sub PS_draw_top_line
{
    my ($page_block, $scale,
	$loc, $length, $xstart, $ystart) = @_;
    my ($xpage_start, $xpage_end, $ypage, $i);
    my ($junk);

    # figure out which pages to put it on
    if ($scale eq 'p') # scaled to one page
    {
	$xpage_start = 0;
	$xpage_end = 0;
	$ypage = 0;
    }
    else # must be 1, 2, 4, 8, or 16
    {
	$xpage_start = (($xstart + 1) / (64 / $scale));
	($xpage_start, $junk) = split(/\./, $xpage_start, 2);

	$xpage_end = (($xstart + $length + 2) / (64 / $scale));
	($xpage_end, $junk) = split(/\./, $xpage_end, 2);

	$ypage = (($ystart) / (80 / $scale));
	($ypage, $junk) = split(/\./, $ypage, 2);
    }

    for ($i = $xpage_start; $i <= $xpage_end; $i++)
    {
	# insert call into page block
	$$page_block{$i, $ypage}
	=~ s/$i $ypage XYPage/$i $ypage XYPage\n  \($loc\) $length $xstart $ystart DrawTopLine/;
    }
}

#
# Subroutine: PS_draw_bottom_line($page_block, $scale, $length, $xstart, 
#                                 $ystart)
# Arguments: $page_block      : reference to associative array of page blocks
#            $scale           : scale factor for graph
#            $length          : x length (goes from right to left)
#            $xstart          : x starting position
#            $ystart          : y starting position
#            
# Return Values: 
#                
# Globals Used: 
#               
# Required Files: 
#                 
# Author: Jason Evans
#
# Description: Inserts PS code to draw a bottom line.
#              
sub PS_draw_bottom_line
{
    my ($page_block, $scale, 
	$length, $xstart, $ystart) = @_;
    my ($xpage_start, $xpage_end, $ypage, $i);
    my ($junk);

    # figure out which page to put it on
    if ($scale eq 'p') # scaled to one page
    {
	$xpage_start = 0;
	$xpage_end = 0;
	$ypage = 0;
    }
    else # must be 1, 2, 4, 8, or 16
    {
	$xpage_start = (($xstart + 1) / (64 / $scale));
	($xpage_start, $junk) = split(/\./, $xpage_start, 2);

	$xpage_end = (($xstart - $length + 1) / (64 / $scale));
	($xpage_end, $junk) = split(/\./, $xpage_end, 2);

	$ypage = (($ystart) / (80 / $scale));
	($ypage, $junk) = split(/\./, $ypage, 2);
    }

    for ($i = $xpage_end; $i <= $xpage_start; $i++)
    {
	# insert call into page block	
	$$page_block{$i, $ypage}
	=~ s/$i $ypage XYPage/$i $ypage XYPage\n  $length $xstart $ystart DrawBottomLine/;
    }
}
    
#
# Subroutine: PS_draw_vert_line($page_block, $scale, $length, $dotted, 
#                               $xstart, $ystart)
# Arguments: $page_block      : reference to associative array of page blocks
#            $scale           : scale factor for graph
#            $length          : vertical length of line
#            $dotted          : 0 if not dotted, 1 if dotted
#            $xstart          : x starting position
#            $ystart          : y starting position
#            
# Return Values: 
#                
# Globals Used: 
#               
# Required Files: 
#                 
# Author: Jason Evans
#
# Description: Inserts PS code to draw a vertical line.
#              
sub PS_draw_vert_line
{
    my ($page_block, $scale,
	$length, $dotted, $xstart, $ystart) = @_;
    my ($xpage, $ypage_start, $ypage_end, $i);
    my ($junk);

    # figure out which pages to put it on
    if ($scale eq 'p') # scaled to one page
    {
	$xpage = 0;
	$ypage_start = 0;
	$ypage_end = 0;
    }
    else # must be 1, 2, 4, 8, or 16
    {
	$xpage = (($xstart + 1) / (64 / $scale));
	($xpage, $junk) = split(/\./, $xpage, 2);

	$ypage_start = (($ystart) / (80 / $scale));
	($ypage_start, $junk) = split(/\./, $ypage_start, 2);

	$ypage_end = (($ystart + $length + 1) / (80 / $scale));
	($ypage_end, $junk) = split(/\./, $ypage_end, 2);
    }

    for ($i = $ypage_start; $i <= $ypage_end; $i++)
    {
	# insert call into page block
	$$page_block{$xpage, $i}
	=~ s/$xpage $i XYPage/$xpage $i XYPage\n  $length $dotted $xstart $ystart DrawVertLine/;
    }
}
    
1;
