#!/usr/local/bin/perl -w

#
# Subroutine: verify_cclogic($fname, $stack_item_size, @Stack)
#
# Arguments: $fname           : Input filename.
#            $stack_item_size : Number of items in a sublist on the stack.
#            @Stack : the stack , containing all of the lines of cc logic.
#                     Each sublist on the stack has 6 items:
#                     position 0 : CCD directive
#                              1 : expression
#                              2 : line number
#                              3 : total lines of code so far
#                              4 : total V(g) so far
#                              5 : total V'(g) so far
#
#                     Note : Only position 0 is important here.
#
# Return Values: None.
#                
# Globals Used: 
#               
# Required Files: 
#                 
# Description: looks for the following errors:
#                  missing #endif
#                  unmatched #endif
#                  #else or #elif outside any if block
#                  #elif following an #else
#              
#              if an error is found, then the function
#              describes the error and dies.
#
sub verify_cclogic{
    my $IF = "#if";
    my $IFD = "#ifdef";
    my $IFND = "#ifndef";
    my $END = "#endif";
    my $ELS = "#else";
    my $ELI = "#elif";
    my $errmsg;
    my $nest_level;
    my @sublist = ();
    my ($fname,$stack_item_size,@stack) = @_;
    my $else_flag;

    $else_flag = 0;
    $nest_level = 0;

    while($#stack != -1){ #While there are lines of CC logic on the stack

	#get a sublist off top (beginning) of the stack
	@sublist = splice(@stack,0,$stack_item_size);	    
	$_ = $sublist[0];

	#Figure out what the CC directive is...
	if (/$IF|$IFD|$IFND/){
	    #got an #if type
	    $else_flag = 0;
	    $nest_level++;
	}
	elsif (/$END/){
	    #got a #endif
	    $else_flag = 0;
	    if ($nest_level == 0){
		#error !
		$errmsg = "\nFatal Error - Unmatched \#endif at line $sublist[2]\n\n";
		die ($errmsg);
	    }
	    $nest_level--;
	}
	elsif (/$ELS/){
	    #got an #else
	    if ($nest_level == 0){
		#error !
		$errmsg = "\nFatal Error - #else statement outside any #if block at line " 
		    . $sublist[2] . "\n\n";
		die ($errmsg);
	    }
	    #set else_flag
	    $else_flag = 1;
	}
	elsif (/$ELI/){
	    #got an #elif
	    if ($nest_level == 0){
		#error !
		$errmsg = "\nFatal Error - \#elif statement outside any \#if block at line " 
		    . $sublist[2] . "\n\n";
		die ($errmsg);
	    }
	    if ($else_flag){
		#error !
		$errmsg = "\nFatal Error - \#elif following an \#else at line " 
		    . $sublist[2] . "\n\n";
		die ($errmsg);
	    }
	    $else_flag = 0;
	}
    }

    #Check to see if all nesting levels have been closed out
    if ($nest_level != 0){
	$errmsg = "\nFatal Error - Missing \#endif\n\n";
	die ($errmsg);
    }
}

1;


