// -*- C++ -*-

#include <stdlib.h>

template <typename Key>
extern unsigned int hash (Key k);

template <>
inline unsigned int hash (void * v) {
  return (unsigned int) ((unsigned long) v);
}

template <>
inline unsigned int hash (const void * v) {
  return (unsigned int) ((unsigned long) v);
}

template <>
inline unsigned int hash (int v) {
  return (unsigned int) v;
}

template <typename Key,
	  typename Value,
	  class Allocator>
class MyHashMap {

public:

  MyHashMap (int size = INITIAL_NUM_BINS)
    : num_bins (size),
      elements (0)
  {
    bins = new (alloc.malloc (sizeof(ListNodePtr) * num_bins)) ListNodePtr;
    for (int i = 0 ; i < num_bins; i++) {
      bins[i] = NULL;
    }
  }

  void set (Key k, Value v) {
    int binIndex = (unsigned int) hash(k) % num_bins;
    ListNode * l = bins[binIndex];
    while (l != NULL) {
      if (l->key == k) {
	l->value = v;
	return;
      }
      l = l->next;
    }
    // Didn't find it.
    insert (k, v);
  }

  Value get (Key k) {
    int binIndex = hash(k) % num_bins;
    ListNode * l = bins[binIndex];
    while (l != NULL) {
      if (l->key == k) {
	return l->value;
      }
      l = l->next;
    }
    // Didn't find it.
    return 0;
  }

  void erase (Key k) {
    int binIndex = hash(k) % num_bins;
    ListNode * l = bins[binIndex];
    ListNode * p = NULL;
    while (l != NULL) {
      if (l->key == k) {
	if (p) {
	  p->next = p->next->next;
	  alloc.free (l);
	} else {
	  bins[binIndex] = NULL;
	}
	elements--;
      }
      p = l;
      l = l->next;
    }
  }


private:

  void insert (Key k, Value v) {
    int binIndex = hash(k) % num_bins;
    ListNode * l = new (alloc.malloc (sizeof(ListNode))) ListNode;
    l->key = k;
    l->value = v;
    l->next = bins[binIndex];
    bins[binIndex] = l;
    elements++;
  }

  enum { INITIAL_NUM_BINS = 511 };

  class ListNode {
  public:
    ListNode (void)
      : next (NULL)
    {}
    Key key;
    Value value;
    ListNode * next;
  };

  int num_bins;
  int elements;

  typedef ListNode * ListNodePtr;
  ListNodePtr * bins;
  Allocator alloc;
};
